from .amplicon_coverage import amplicon_coverage, gc_depth
from .analyze import find_mutants
from .cmds import aa, nt
from .lineages import find_lineages
# from .consensus import consensus_from_bam


class CovBreakdown(object):
    """Identify frequencies of concerning mutations from aligned reads"""

    def __str__(self):
        return "Identify frequencies of concerning mutations from aligned reads"

    def aa(self, mut):
        aa(mut)

    def nt(self, mut):
        nt(mut)

    def find_mutants(self, samples_path, mutations_path=None, min_depth=40, save_img=False, csv=False):
        find_mutants(samples_path, mutations_path=mutations_path, min_depth=min_depth, save_img=save_img, csv=csv)

    def find_lineages(self, samples_path, lineages_path=None, ts=False, csv=False, min_depth=40, show_stacked=False, unique=False, save_img=False, l2=False):
        find_lineages(samples_path, lineages_path=lineages_path, ts=ts, csv=csv, min_depth=min_depth, show_stacked=show_stacked, unique=unique, save_img=save_img, l2=l2)

    def amplicon_coverage(self, samples_path):
        amplicon_coverage(samples_path)

    def gc_depth(self, samples_path):
        gc_depth(samples_path)



# -*- coding: utf-8 -*-
"""
Created on Wed May 31 2023

@author: Jenn Knapp
email: jknapp@uwaterloo.ca
"""
"""
Purpose: Retrieve lineage-definining mutations from Cov-Spectrum for all lineages that have deposited sequences

Reguires:
lineages.txt (All lineages listed when searching CoV-Spectrum with the params World, AllTimes, and blank search bar)

"""
import requests
import json

# Download master lineage list from cov-lineages pango-designation (updated regularly as new lineages arise)
input_url = "https://raw.githubusercontent.com/cov-lineages/pango-designation/master/lineage_notes.txt"
output_file = "lineages.txt"

response = requests.get(input_url)
lines = response.text.splitlines()

# Parse lineage list to exclude header, descriptions, and withdrawn lineages 
values = [line.split()[0] for line in lines[1:] if not line.startswith("*")]

# Write all lineage names to a file called lineages.txt
with open(output_file, "w") as file:
    file.write("\n".join(values))
    file.close()  # Close the file to ensure data is written to disk

base_url = "https://lapis.cov-spectrum.org/open/v1"

# Read lineage list
with open("lineages.txt", "r") as file:
    lineages = file.read().splitlines()

# Iterate over lineages
for lineage in lineages:
    output_file = "data/constellations/" + lineage + ".json"

# Construct URL
    url1 = base_url + "/sample/aggregated?nextcladePangoLineage=" + lineage

    try:
        # Send API request
        response = requests.get(url1)

        # Check if the request was successful (status code 200)
        if response.status_code == 200:
            # Parse the response as JSON
            data1 = response.json()

            # Extract the required data from the JSON response
            count = data1["data"][0]["count"]

            if count > 100:

                # Construct URL
                url2 = base_url + "/sample/nuc-mutations?nextcladePangoLineage=" + lineage

                try:
                    # Send API request
                    response = requests.get(url2)

                    # Check if the request was successful (status code 200)
                    if response.status_code == 200:
                        # Parse the response as JSON
                        data = response.json()

                        # Extract the required data from the JSON response
                        nucleotide_mutations = [
                            mutation["mutation"] for mutation in data["data"]
                            if mutation["proportion"] > 0.90 and not mutation["mutation"].endswith("-")
                        ]

                        # Create the final JSON structure
                        json_data = {
                            "label": lineage + "-like",
                            "description": f"{lineage} lineage defining mutations",
                            "sources": [],
                            "tags": [lineage],
                            "sites": nucleotide_mutations,
                            "note": "Unique mutations for sublineage"
                        }

                        # Write data to JSON file
                        with open(output_file, "w") as file:
                            json.dump(json_data, file)
                    else:
                        print("Error occurred while fetching data for lineage:", lineage)
         
                except requests.RequestException as e:
                        print("An error occurred during the API request:", str(e))

    except requests.RequestException as e:
        print("An error occurred during the API request:", str(e))




import json

from Bio import SeqIO


def process_reference():
    cov = list(SeqIO.parse("sequence.gb", "genbank"))[0]

    genes = {}

    for f in cov.features:
        if f.type == 'gene':
            gene = f.qualifiers['gene'][0]
            start = int(f.location.start)
            end = int(f.location.end)
            genes[gene] = [start, end]

    with open('sars_cov_2.py', 'w') as f:
        f.write('genes = {}\n\nseq = \'{}\''.format(genes, cov.seq))


def get_amplicons():
    cov = list(SeqIO.parse("sequence.gb", "genbank"))[0]
    with open('nCoV-2019.insert.bed', 'r') as f:
        inserts = [l.split('\t') for l in f.read().split('\n')[:-1]]

    for i in range(len(inserts)):
        insert = inserts[i]
        section = cov.seq[int(insert[1]):int(insert[2])]
        gc = (section.count('G') + section.count('C')) / len(section)
        inserts[i].append(gc)

    with open('artic_amplicons.py', 'w') as f:
        f.write('inserts = {}'.format(inserts))


def fix_mut_name(old_mut_name):
    mut_name = old_mut_name.upper()
    if '/' in mut_name:
        # If multiple amino acid dels, only take the first
        mut_name = mut_name[:mut_name.find('/')]
    if not mut_name.startswith('ORF'):
        return mut_name
    orf = 'ORF'
    col_idx = mut_name.find(':')
    mid = mut_name[3:col_idx]
    end = mut_name[col_idx:]
    return orf + mid.lower() + end


def get_mutations():
    # TODO figure out capitalization
    with open('mutations.json', 'r') as f:
        raw_mutations = json.loads(f.read())

    muts = list(set([fix_mut_name(m['mutation']) for m in raw_mutations]))
    lins = list(set([m['pangolin_lineage'].upper() for m in raw_mutations]))
    mut_lins = {mut: {lin: 0 for lin in lins} for mut in muts}
    for raw_m in raw_mutations:
        mut = fix_mut_name(raw_m['mutation'])
        lin = raw_m['pangolin_lineage'].upper()
        prev = raw_m['prevalence']
        mut_lins[mut][lin] = prev

    with open('mutations.py', 'w') as f:
         f.write('mutations = {}'.format(mut_lins))


def get_who_mutations():
    with open('clusters.json', 'r') as f:
        lineages = json.loads(f.read())['clusters']
    vocs = []
    for lin in lineages:
        if 'who_name' in lin and len(lin['who_name']) > 0:
            vocs.append(lin)
    mutations = {}
    who_names = [voc['who_name'][0] for voc in vocs] + ['BA.1', 'BA.2']
    print(who_names)
    who_names = [wn for wn in who_names if wn != 'Omicron']
    print(who_names)
    for voc in vocs:
        who_name = voc['who_name'][0]
        adn = 'alt_display_name'
        if adn in voc and len(voc[adn]) and voc[adn][0] == 'BA.2':
            who_name = 'BA.2'
        # if adn in voc and len(voc[adn]) and voc[adn][0] == 'BA.1':
        if who_name == 'Omicron':
            who_name = 'BA.1'
        muts = voc['mutations']['nonsynonymous']
        muts += voc['mutations']['synonymous']
        for mut in muts:
            if 'gene' in mut:
                if mut['gene'] == 'ORF9b':
                    continue
                if mut['right'] == '-':
                    mut_name = '{}:DEL{}'.format(mut['gene'], mut['pos'])
                else:
                    mut_name = '{}:{}{}{}'.format(
                        mut['gene'],
                        mut['left'],
                        mut['pos'],
                        mut['right']
                    )
            else:
                mut_name = '{}{}{}'.format(
                    mut['left'],
                    mut['pos'],
                    mut['right']
                )
            # TODO: mutations after deletions don't work
            blacklist = [
                'ORF1a:P2287S',
                # Omicron post-deletion
                'ORF1a:L2084I',
                'ORF1a:I3758V',
                'S:Y145D',
                'S:L212I',
                # BA.2 post-deletion
                'S:A27S',
                # Possibly in Omicron
                'ORF1a:DEL3677',
                'S:R346K',
                # Delta post-deletion
                'S:R158G'
            ]
            if mut_name in blacklist:
                continue
            if mut_name not in mutations:
                mutations[mut_name] = {who_name: 0 for who_name in who_names}
            mutations[mut_name][who_name] = 1

    with open('mutations.py', 'w') as f:
         f.write('mutations = {}'.format(mutations))


def get_constellations():
    import json
    from os import getenv, listdir, mkdir, system
    from os.path import isfile, join

    def parse_mut(mut):
        mut = mut.upper()
        if mut[0].isnumeric():
            mut = 'ORF' + mut
        mut = mut.replace('SPIKE', 'S')
        mut = mut.replace('NUC:', '')
        mut = mut.replace('ORF1AB', 'ORF1ab')
        mut = mut.replace('ORF1A', 'ORF1a')
        mut = mut.replace('ORF1B', 'ORF1b')
        mut = mut.replace('ORF3A', 'ORF3a')
        mut = mut.replace('ORF7A', 'ORF7a')
        mut = mut.replace('ORF7b', 'ORF7b')
        return mut

    data_paths = [
        './data/constellations/',
#        './data/extra_constellations/',
    ]

    constellations = {}
    mutations = {}
    for data_path in data_paths:
        fns = [f for f in listdir(data_path) if isfile(join(data_path, f)) and f.endswith('.json')]
        fns.sort()
        for fn in fns:
            with open('{}/{}'.format(data_path, fn), 'r') as f:
                voc = json.loads(f.read())
            label = voc['label']
            constellations[label] = [parse_mut(mut) for mut in voc['sites']]
    with open('constellations.py', 'w') as f:
         f.write('constellations = {}'.format(constellations))
    vocs = list(constellations.keys())
    # Remove non-specific lineages
#    constellations['Omicron (XE-like)'] += constellations['XE-parent1']
#    constellations['Omicron (XE-like)'] += constellations['XE-parent2']
#    for voc in constellations:
#        if 'Omicron' in voc:
#            constellations[voc] += constellations['Omicron (Unassigned)']
#    vocs.remove('Omicron (Unassigned)')
#    vocs.remove('XE-parent1')
#    vocs.remove('XE-parent2')
    for voc in vocs:
        for mut_name in constellations[voc]:
            # if mut_name in blacklist:
            #     continue
            if '+' in mut_name: # TODO: support
                continue
            if mut_name.startswith('NSP'): # TODO: support
                continue
            if mut_name not in mutations:
                mutations[mut_name] = {voc: 0 for voc in vocs}
            mutations[mut_name][voc] = 1
    with open('mutations.py', 'w') as f:
         f.write('mutations = {}'.format(mutations))


if __name__ == '__main__':
    # process_reference()
    # get_amplicons()
    # get_mutations()
    # get_who_mutations()
    get_constellations()



from collections import defaultdict
from math import ceil, floor

from .analyze import find_mutants_in_bam, print_mut_results
from .convert_mutations import aa, nt
from .mutations import mutations as mut_lins


def parse_mutations(mutations):
    nts = [mut for mut in mutations if ':' not in mut]
    aas = [mut for mut in mutations if ':' in mut]
    return nts + sum([aa(mut) for mut in aas], [])


def parse_snv(snv):
    pos = int(snv[1:-1])
    old_bp = snv[0]
    new_bp = snv[-1]
    return old_bp, pos, new_bp


def mut_in_col(pileupcolumn, mut):
    muts = 0
    not_muts = 0
    if mut == '-':
        for pileupread in pileupcolumn.pileups:
            if pileupread.is_del:
                muts += 1
            else:
                not_muts += 1
    else:
        for pileupread in pileupcolumn.pileups:
            qpos = pileupread.query_position
            if qpos is None:
                continue
            base = pileupread.alignment.query_sequence[qpos]
            if base == mut:
                muts += 1
            else:
                not_muts += 1
    return muts, not_muts


def write_csv(sample_results, sample_names, home_lab):
    lin_names = set()
    for sr in sample_results:
        for key in sr.keys():
            lin_names.add(key)
    lin_names = [n for n in lin_names]
    lin_names.sort()
    csv_headers = ['Sample name'] + [n for n in lin_names]
    csv_rows = []
    for i in range(len(sample_names)):
        sr = sample_results[i]
        csv_rows.append([sample_names[i]] + [str(round(sr[n], 3)) if n in sr else '0' for n in lin_names])
    with open('{}_lineages.csv'.format(home_lab), 'w') as f:
        f.write('\n'.join(','.join(row) for row in [csv_headers] + csv_rows))


def plot_lineages(sample_results, sample_names, img_path=None, all_lins=False):
    import numpy as np
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    names = set()
    for sr in sample_results:
        for key in sr.keys():
            if sr[key] > 0.01 or all_lins:
                names.add(key)
    names = [n for n in names]
    names.sort()
    # names = sample_results[0].keys()
    lin_fractions = np.array([[lin_results[lin]*100 if lin in lin_results else 0 for lin in names] for lin_results in sample_results]).T    
    # no_reads = np.array([[f == -1 for f in fractions] for fractions in lin_fractions])
    # fig, ax = plt.subplots(figsize=(len(sample_names)/2,len(names)/2))
 
# get the tick label font size
    fontsize_pt = plt.rcParams['ytick.labelsize']
    dpi = 72.27  
# comput the matrix height in points and inches
    matrix_height_pt = fontsize_pt * (len(lin_fractions)+30)
    matrix_height_in = matrix_height_pt / dpi
# compute the required figure height 
    top_margin = 0.10  # in percentage of the figure height
    bottom_margin = 0.20 # in percentage of the figure height
    figure_height = matrix_height_in / (1 - top_margin - bottom_margin)
    figure_width = len(sample_names) * 2 + 5
# build the figure instance with the desired height
    fig, ax = plt.subplots(
        figsize=(figure_width, figure_height), 
        gridspec_kw=dict(top=1-top_margin, bottom=bottom_margin),
)
    ax = sns.heatmap(
        lin_fractions,
        annot=True,
    #   annot_kws={'fontsize':10},
    #   mask=no_reads,
        cmap=sns.cm.rocket_r,
        xticklabels=sample_names,
        yticklabels=names,
        vmin=0,
        vmax=100,
    #   square=True,
        cbar_kws={'format': '%.0f%%'},
        fmt='.1f',
    )
    plt.xlabel('Frequency in sample')
    plt.xticks(rotation=30, ha="right", rotation_mode="anchor")
    plt.ylabel('SARS-CoV-2 Lineage')
    # plt.tight_layout()
    if img_path is not None:
        plt.subplots_adjust(bottom=0.3, left=0.6)
        plt.savefig(img_path, dpi=300)
    else:
        plt.show()
    
def plot_lineages_timeseries(sample_results, sample_names):
    from datetime import date
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()

    names = set()
    for sr in sample_results:
        for key in sr.keys():
            names.add(key)
    names = [n for n in names]
    names.sort()
    # names = list(sample_results[0].keys())
    # names.sort()
    if '_' in sample_names[0]:
        locations = [dt.split('_')[0] for dt in sample_names]
    # loc_data = {}
    locations = []
    dates = []
    for i in range(len(sample_names)):
        loc, dt = sample_names[i].split('_')
        dt = date.fromisoformat(dt)
        locations.append(loc)
        dates.append(dt)
    loc_set = sorted(list(set(locations)))
    ncols = 3
    nrows = ceil(len(loc_set) / ncols)
    fig, axes = plt.subplots(nrows, ncols)
    plt.tight_layout()
    non_zero = set()
    for sr in sample_results:
        for name in sr:
            if sr[name] > 0:
                non_zero.add(name)
    for sr in sample_results:
        for name in list(sr.keys()):
            if name not in non_zero:
                del sr[name]
    names = sorted(list(non_zero))
    for r in range(nrows):
        for c in range(ncols):
            i = r * ncols + c
            if i < len(loc_set):
                loc = loc_set[i]
                loc_results = []
                loc_dates = []
                for i in range(len(sample_results)):
                    sample_loc = locations[i]
                    if sample_loc == loc:
                        loc_results.append(sample_results[i])
                        loc_dates.append(dates[i])
                d = {name: [sr[name] if name in sr else 0 for sr in loc_results] for name in names}
                df = pd.DataFrame(data=d, index=loc_dates)
                # df = df.loc[:, df.any()] # Delete all-zero lineages
                axes[r,c].set_title(loc)
                axes[r,c].set_xlim([min(dates),max(dates)])
                axes[r,c].set_ylim([0,1])
                df.plot.area(ax=axes[r,c], fontsize=8, legend=False, rot=30)
            else:
                fig.delaxes(axes[r,c])
    leg = fig.legend(labels=names, loc="lower right", ncol=3)
    for line in leg.get_lines():
        line.set_linewidth(6)
    plt.show()


def show_lineage_predictions(sample_results, X, Y, covered_muts):
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    merged_lins = list(sample_results.keys())
    d = {ml: [] for ml in merged_lins}
    d['Observed'] = []
    idx = []
    for i in range(len(covered_muts)):
        Y_p = 0
        for j in range(len(merged_lins)):
            ml = merged_lins[j]
            lin_contribution = X[i][j]*sample_results[ml]
            Y_p += lin_contribution
        if Y[i] < 0.02 and Y_p < 0.02:
            # Skip non-present mutations
            # pass
            continue
        cm = covered_muts[i]
        idx.append('{}\nobserved'.format(cm))
        idx.append('{}\npredicted'.format(cm))
        for j in range(len(merged_lins)):
            ml = merged_lins[j]
            lin_contribution = X[i][j]*sample_results[ml]
            d[ml] += [0, lin_contribution]
        d['Observed'] += [Y[i], 0]
    df = pd.DataFrame(data=d, index=idx)
    df = df.loc[:, df.any()] # Delete all-zero lineages
    if len(df) == 0:
        print('No mutations to show')
        return
    df.plot.bar(stacked=True, rot=0)
    plt.tight_layout()
    plt.show()


def show_lineage_pie(sample_results):
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    merged_lins = list(sample_results.keys())
    freqs = [sample_results[ml] for ml in merged_lins]
    if sum(freqs) < 1:
        # merged_lins.append('Other')
        # freqs.append(1 - sum(freqs))
        freqs = [f/sum(freqs) for f in freqs]
    df = pd.DataFrame(data={'Fraction': freqs}, index=merged_lins)
    df = df[df.Fraction > 0] # Delete all-zero lineages
    print(df)
    print(sample_results)
    df.plot.pie(y='Fraction', legend=False, autopct='%1.1f%%', ylabel='')
    plt.show()


def do_regression(lmps, Y):
    # Perform linear regression and redo if sum(frequencies) > 1
    import numpy as np
    from sklearn.linear_model import LinearRegression
    from sklearn.linear_model import Lasso
    X = np.array(lmps).T
    reg = LinearRegression(fit_intercept=0, positive=True).fit(X, Y)
    score = reg.score(X, Y)
    best_score = score
    best_reg = reg
    if sum(reg.coef_) > 1:
        # If there is no valid solution, drop uncovered mutations one at a time
        # This allows for lineages with one missing mutation
        valid = False
        best_score = 0.8 # minimum valid score
        for i in range(len(X)):
            if Y[i] == 0:
                new_X = np.concatenate((X[:i], X[i+1:]))
                new_Y = np.concatenate((Y[:i], Y[i+1:]))
                new_reg = LinearRegression(fit_intercept=0, positive=True).fit(new_X, new_Y)
                new_score = new_reg.score(X, Y)
                if sum(new_reg.coef_) <= 1 and new_score > best_score:
                    valid = True
                    best_score = new_score
                    best_reg = new_reg
        if not valid:
            print('Warning: solutions sums to > 1')
    return X, [frac for frac in best_reg.coef_]


def do_regression_linear(lmps, Y, muts):
    # Linear program for minimizing error of frequencies
    import numpy as np
    # from sklearn.linear_model import LinearRegression
    # from sklearn.linear_model import Lasso
    from ortools.linear_solver import pywraplp
    from ortools.init import pywrapinit
    solver = pywraplp.Solver.CreateSolver('GLOP')
    num_lins = len(lmps)
    num_muts = len(lmps[0])
    lins = [solver.NumVar(0, 1, 'x_{}'.format(i)) for i in range(num_lins)]
    t = [solver.NumVar(0, solver.infinity(), 't_{}'.format(i)) for i in range(num_muts)]
    # t = [solver.NumVar(0, 1, 't_{}'.format(i)) for i in range(num_muts)]
    # t = [solver.NumVar(-1, 1, 't_{}'.format(i)) for i in range(num_muts)]
    constraints_1 = []
    constraints_2 = []
    # print(lmps)
    mut_freqs = list(Y)
    # print(mut_freqs)
    for j in range(num_muts):
        constraints_1.append(solver.Constraint(-solver.infinity(), mut_freqs[j], 'c_{}_1'.format(j)))
        constraints_2.append(solver.Constraint(mut_freqs[j], solver.infinity(), 'c_{}_2'.format(j)))
        constraints_1[j].SetCoefficient(t[j], -1)
        constraints_2[j].SetCoefficient(t[j], 1)
        for i in range(num_lins):
            constraints_1[j].SetCoefficient(lins[i], lmps[i][j])
            constraints_2[j].SetCoefficient(lins[i], lmps[i][j])
    # freqs = solver.Constraint(0, solver.infinity(), 'frequencies')
    freqs = solver.Constraint(0, 1, 'frequencies')
    for i in range(num_lins):
        freqs.SetCoefficient(lins[i], 1)
    objective = solver.Objective()
    for j in range(num_muts):
        objective.SetCoefficient(t[j], 1)
    objective.SetMinimization()
    status = solver.Solve()
    # print('Lineage abundances:')
    # for i in range(num_lins):
    #     print(lins[i].solution_value())
    print('Residuals:')
    sig_idxs = []
    diffs = []
    mut_diffs = {}
    for j in range(num_muts):
        res = t[j].solution_value()
        yp = sum([lins[i].solution_value() * lmps[i][j] for i in range(num_lins)])
        diffs.append(Y[j] - yp)
        mut_diffs[muts[j]] = diffs[j]
        if res > 0.1:
            print(muts[j])
            print(res)
            print(diffs[j])
            sig_idxs.append(j)
    if len(sig_idxs) > 0 and False:
        d = {}
        d['muts'] = [muts[i] for i in sig_idxs]
        d['residuals'] = [diffs[i] for i in sig_idxs]
        # d = {muts[i]: t[i].solution_value() for i in range(num_muts)}
        import matplotlib.pyplot as plt
        import pandas as pd
        df = pd.DataFrame(data=d)
        # df = df.loc[:, df.any()] # Delete all-zero lineages
        df.plot.bar(x='muts', y='residuals')
        plt.tight_layout()
        plt.show()
    X = np.array(lmps).T

    return X, [lin.solution_value() for lin in lins], mut_diffs


def find_lineages_in_bam(bam_path, return_data=False, min_depth=40, lineages=[], unique=False, l2=False):
    import numpy as np
    import pysam

    samfile = pysam.Samfile(bam_path, "rb")

    aa_mutations = [m for m in mut_lins.keys()]
    # aa_mutations = [m for m in mut_lins.keys() if m[0] in ['S']] # Only spike
    # aa_mutations = [m for m in mut_lins.keys() if m[0] in ['N']] # Only N
    aa_blacklist = ['S:D614G'] # all lineages contain this now
    aa_mutations = [m for m in aa_mutations if m not in aa_blacklist]
    # lineages = ['Delta', 'BA.1']
    if len(lineages) == 0:
        lineages = list(mut_lins['C241T'].keys()) # arbitrary
    if unique:
        aa_mutations = [mut for mut in aa_mutations if sum(mut_lins[mut][l] for l in lineages) == 1]
    mutations = parse_mutations(aa_mutations)
    vocs = ['B.1.1.7', 'B.1.617.2', 'P.1', 'B.1.351']
    vois = ['B.1.525', 'B.1.526', 'B.1.617.1', 'C.37']
    # if only_vocs:
    # lineages = vocs + vois
    # lineages = ['Omicron', 'BA.2', 'Delta']

    mut_results = find_mutants_in_bam(bam_path, aa_mutations)

    covered_muts = [m for m in aa_mutations if sum(mut_results[m]) >= min_depth]
    if len(covered_muts) == 0:
        print('No coverage')
        return None
    covered_lineages = set()
    for m in covered_muts:
        for l in mut_lins[m]:
            if mut_results[m][0] > 0 and mut_lins[m][l] > 0.5:
                covered_lineages.add(l)
    covered_lineages = [l for l in lineages if l in covered_lineages]
    Y = np.array([mut_results[m][0]/sum(mut_results[m]) if sum(mut_results[m]) > 0 else 0 for m in covered_muts])
    lin_mut_profiles = [[round(mut_lins[mut][lin]) for mut in covered_muts] for lin in lineages]
    # Merge indistinguishable lineages
    merged_lmps = []
    merged_lins = []
    for i in range(len(lineages)):
        lmp = lin_mut_profiles[i]
        lin = lineages[i]
        if lmp in merged_lmps:
            lmp_idx = merged_lmps.index(lmp)
            merged_lins[lmp_idx] += ' or '
            merged_lins[lmp_idx] += lin
        else:
            merged_lmps.append(lmp)
            merged_lins.append(lin)
    if l2:
        X, reg = do_regression(merged_lmps, Y)
    else:
        X, reg, mut_diffs = do_regression_linear(merged_lmps, Y, covered_muts)

    # print_mut_results(mut_results)
    sample_results = {merged_lins[i]: round(reg[i], 3) for i in range(len(merged_lins))}
    print(sample_results)
    # Normalize frequencies
    # freq_sum = sum(sample_results[lin] for lin in sample_results)
    # for lin in sample_results:
    #     sample_results[lin] = sample_results[lin] / freq_sum

    if return_data:
        return sample_results, X, Y, covered_muts
    # TODO: rethink information flow
    # return sample_results, mut_diffs
    return sample_results


def find_lineages(file_path, lineages_path, ts, csv, min_depth, show_stacked, unique, save_img, l2):
    """
    Accepts either a bam file or a tab delimited  txt file like
    s1.bam  Sample 1
    s2.bam  Sample 2
    """

    sample_results = []
    sample_mut_diffs = defaultdict(list)
    sample_names = []
    home_lab = file_path.replace('.txt','')
    lineages = []
    all_lins = False # Show all lineages
    if lineages_path is not None:
        all_lins = True
        with open(lineages_path, 'r') as f:
            lineages = f.read().splitlines()
    if file_path.endswith('.bam'):
        sr, X, Y, covered_muts = find_lineages_in_bam(file_path, True, min_depth, lineages, unique, l2)
        if show_stacked:
            show_lineage_predictions(sr, X, Y, covered_muts)
            show_lineage_pie(sr)
        sample_results.append(sr)
        sample_names.append('')
    else:
        with open(file_path, 'r') as f:
            samples = [line.split('\t') for line in f.read().split('\n')]
        for sample in samples:
            if sample[0].endswith('.bam'): # Mostly for filtering empty
                print('{}:'.format(sample[1]))
                # sample_result, mut_diffs = find_lineages_in_bam(sample[0], False, min_depth, lineages, unique)
                sample_result = find_lineages_in_bam(sample[0], False, min_depth, lineages, unique, l2)
                if sample_result is not None and sum(sample_result.values()) > 0:
                    sample_results.append(sample_result)
                    sample_names.append(sample[1])
                    # for mut in mut_diffs:
                    #     sample_mut_diffs[mut].append(mut_diffs[mut])
                # sample_results.append(sample_result)
                # sample_names.append(sample[1])
                print()
    # print(sample_mut_diffs)
    for mut in sample_mut_diffs:
        diffs = sample_mut_diffs[mut]
        if abs(sum(diffs)/len(diffs)) > 0.1:
            print(mut)
            print(diffs)
    img_path = file_path.replace('.txt', '.png') if save_img else None
    if ts:
        plot_lineages_timeseries(sample_results, sample_names)
    else:
        plot_lineages(sample_results, sample_names, img_path, all_lins)
    if csv:
        write_csv(sample_results, sample_names, home_lab)



from .artic_amplicons import inserts
from .convert_mutations import aa, nt


# inserts = inserts[1:98]

def plot_depths(sample_results, sample_names):
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    import seaborn as sns
    # sns.set_theme(style="whitegrid")
    # samples = sum([98*[name] for name in sample_names], [])
    # samples = [s[71:85] for s in samples]
    samples = sum([len(inserts)*[name] for name in sample_names], [])
    amplicons = sum([list(amplicon.keys()) for amplicon in sample_results], [])
    # pools = ['Pool 1' if int(a) % 2 == 0 else 'Pool 2' for a in amplicons]
    pools = ['Pool 1' if True else 'Pool 2' for a in amplicons]
    depths = sum([[np.log(a+1) for a in amplicon.values()] for amplicon in sample_results], [])
    d = {
        'Sample': samples,
        'Amplicon number': amplicons,
        'Pool': pools,
        'Log depth': depths
    }
    df = pd.DataFrame(data=d)
    g = sns.FacetGrid(df, row="Sample", hue="Pool", height=1.7, aspect=8)
    g.map(sns.barplot, "Amplicon number", "Log depth", order=[str(i) for i in range(1,99)], hue_order=['Pool 1', 'Pool 2'])
    # plt.locator_params(axis='x', nbins=20)
    plt.locator_params(axis='x')
    plt.savefig('SA_coverage.png')
    plt.show()


def plot_depths_gc(sample_results, sample_names):
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    import seaborn as sns
    from scipy import stats
    # sns.set_theme(style="whitegrid")
    depths = sample_results[0]
    samples = sum([98*[name] for name in sample_names], [])
    # for i in range(len(samples)):
    #     samples[i] = 'Test'
    amplicons = sum([list(amplicon.keys()) for amplicon in sample_results], [])
    gcs = [inserts[int(a)-1][6] for a in amplicons]
    pools = ['Pool 1' if int(a) % 2 == 0 else 'Pool 2' for a in amplicons]
    depths = sum([[np.log(a+1) for a in amplicon.values()] for amplicon in sample_results], [])
    d = {
        'Sample': samples,
        'Amplicon number': [int(a) for a in amplicons],
        'GC content': gcs,
        'Pool': pools,
        'Log depth': depths
    }
    df = pd.DataFrame(data=d)
    # df = df.loc[df['Log depth'] != 0]
    g = sns.FacetGrid(df, col="Sample")
    g.map(sns.regplot, "GC content", "Log depth")
    statistic, pvalue = stats.pearsonr(gcs, depths)
    print("Pearson correlation statistic: {}".format(statistic))
    print("Pearson p-value (probability that variables are independant): {}".format(pvalue))
    # print(res.confidence_interval())
    # g.map(sns.regplot, "Amplicon number", "Log depth")
    plt.show()


def plot_amplified_fraction(sample_results, sample_names):
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    import seaborn as sns
    # sns.set_theme(style="whitegrid")
    depths = sample_results[0]
    d = {
        'Sample': sum([98*[name] for name in sample_names], []),
        'Amplicon': sum([list(amplicon.keys()) for amplicon in sample_results], []),
        'Depth': sum([[min(d, 1) for d in amplicon.values()] for amplicon in sample_results], [])
    }
    df = pd.DataFrame(data=d)
    g = sns.FacetGrid(df, row="Sample")
    g.map(sns.barplot, "Amplicon", "Depth")
    plt.locator_params(axis='x', nbins=20)
    plt.show()


def find_depths_in_bam(bam_path, max_depth=50000):
    import pysam

    samfile = pysam.Samfile(bam_path, "rb")

    amp_mids = {int((int(i[1]) + int(i[2])) / 2): i[3] for i in inserts}
    amplified = {i[3]: 0 for i in inserts}

    for pileupcolumn in samfile.pileup(max_depth=max_depth):
        pos = pileupcolumn.pos
        if pos in amp_mids:
            depth = pileupcolumn.get_num_aligned()
            amplified[amp_mids[pos]] = depth
    samfile.close()

    # SAMPLE_NUM = '2'

    # with open('s{}_mixed.csv'.format(SAMPLE_NUM), 'w') as f:
    #     f.write('\n'.join(['{},{}'.format(insert_name, amplified[insert_name]) for insert_name in amplified.keys()]))

    return amplified


def amplicon_coverage(file_path):
    """
    Accepts either a bam file or a tab delimited  txt file like
    s1.bam  Sample 1
    s2.bam  Sample 2
    """
    sample_results = []
    sample_names = []
    if file_path.endswith('.bam'):
        sample_results.append(find_depths_in_bam(file_path))
        sample_names.append('')
    else:
        with open(file_path, 'r') as f:
            samples = [line.split('\t') for line in f.read().split('\n')]
        for sample in samples:
            if sample[0].endswith('.bam'): # Mostly for filtering empty
                sample_results.append(find_depths_in_bam(sample[0]))
                sample_names.append(sample[1])
    plot_depths(sample_results, sample_names)


def gc_depth(file_path):
    """
    Accepts either a bam file or a tab delimited  txt file like
    s1.bam  Sample 1
    s2.bam  Sample 2
    """
    sample_results = []
    sample_names = []
    if file_path.endswith('.bam'):
        sample_results.append(find_depths_in_bam(file_path))
        sample_names.append('')
    else:
        with open(file_path, 'r') as f:
            samples = [line.split('\t') for line in f.read().split('\n')]
        for sample in samples:
            if sample[0].endswith('.bam'): # Mostly for filtering empty
                sample_results.append(find_depths_in_bam(sample[0]))
                sample_names.append(sample[1])
    plot_depths_gc(sample_results, sample_names)



from .convert_mutations import aa, nt
from .mutations import mutations as mut_lins


def parse_snv(snv):
    pos = int(snv[1:-1])
    old_bp = snv[0]
    new_bp = snv[-1]
    return old_bp, pos, new_bp


def snv_name(snv):
    return '{}{}{}'.format(*snv)

def parse_mutation(mut):
    if ':' in mut:
        muts = aa(mut)
    else:
        muts = [mut]
    return [parse_snv(m) for m in muts]


def mut_in_col(pileupcolumn, mut):
    muts = 0
    not_muts = 0
    if mut == '-':
        for pileupread in pileupcolumn.pileups:
            if pileupread.is_del:
                muts += 1
            else:
                not_muts += 1
    else:
        for pileupread in pileupcolumn.pileups:
            qpos = pileupread.query_position
            if qpos is None:
                not_muts += 1
                continue
            base = pileupread.alignment.query_sequence[qpos]
            if base == mut:
                muts += 1
            else:
                not_muts += 1
    return muts, not_muts


def print_mut_results(mut_results, min_depth):
    cov = 0
    mut_cov = 0
    for name in mut_results:
        muts, not_muts = mut_results[name]
        # new_base = name[-1]
        # print('{}:'.format(name))
        total = muts + not_muts
        if total >= min_depth:
            cov += 1
            if muts > 0:
                mut_cov += 1

    # for name in mut_results:
    #     muts, not_muts = mut_results[name]
    #     new_base = name[-1]
    #     print('{}:'.format(name))
    #     total = muts + not_muts
    #     if total == 0:
    #         print('No coverage of {}'.format(name))
    #     else:
    #         print('{} are {}, {} are wildtype ({:.2f}% of {} total)'.format(
    #             muts,
    #             new_base,
    #             not_muts,
    #             muts/total*100,
    #             total
    #         ))

    print('{}/{} mutations covered'.format(cov, len(mut_results)))
    print('{}/{} mutations detected'.format(mut_cov, len(mut_results)))

def plot_mutations(sample_results, sample_names, min_depth, img_path):
    import numpy as np
    import matplotlib.pyplot as plt
    import seaborn as sns; sns.set_theme()
    names = sample_results[0].keys()
#    print(names)
# Print both aa and nt changes as Y-tick labels
#    for i in range(len(names)):
#        mutant = names[i]
#        if ":" in mutant:
#            mutant = mutant + " - " + aa(mutant)
#        else:
#            mutant = nt(mutant) + " - " + mutant
#        names[i] = mutant
#    print(names)
    sample_counts = [[mut_results[mut] for mut in names] for mut_results in sample_results]
    num_mutations = len(names)
    mut_fractions = [[] for _ in range(num_mutations)]
    for i in range(num_mutations):
        for counts in sample_counts:
            count = counts[i]
            total = count[0] + count[1]
            fraction = count[0]/total if total >= min_depth else -1
            # mut_fractions[i].append(round(fraction, 2))
            mut_fractions[i].append(round(fraction, 4))

    no_reads = np.array([[f == -1 for f in fractions] for fractions in mut_fractions])

# get the tick label font size
    fontsize_pt = plt.rcParams['ytick.labelsize']
    dpi = 72.27

# comput the matrix height in points and inches
    matrix_height_pt = fontsize_pt * (len(mut_fractions)+30)
    matrix_height_in = matrix_height_pt / dpi

# compute the required figure height 
    top_margin = 0.10  # in percentage of the figure height
    bottom_margin = 0.20 # in percentage of the figure height
    figure_height = matrix_height_in / (1 - top_margin - bottom_margin)
    figure_width = len(sample_names) * 2 + 5

# build the figure instance with the desired height
    fig, ax = plt.subplots(
        figsize=(figure_width, figure_height), 
        gridspec_kw=dict(top=1-top_margin, bottom=bottom_margin),
)

    ax = sns.heatmap(
        mut_fractions,
        annot=True,
	# annot_kws={'fontsize':10},
        mask=no_reads,
        cmap=sns.cm.rocket_r,
        xticklabels=sample_names,
        yticklabels=names,
        vmin=0,
        vmax=1,
	fmt='.2f',
    )
    plt.xlabel('Sample')
    plt.xticks(rotation=30, ha="right", rotation_mode="anchor")
    plt.ylabel('Mutation')
    # plt.tight_layout()
    if img_path is not None:
        plt.savefig(img_path, dpi=300)
    else:
        plt.show()

def write_csv(sample_results, sample_names, min_depth, home_lab, mutants_name):
    names = sample_results[0].keys()
    sample_counts = [[mut_results[mut] for mut in names] for mut_results in sample_results]
    num_mutations = len(names)
    mut_fractions = [[] for _ in range(num_mutations)]
    for i in range(num_mutations):
        for counts in sample_counts:
            count = counts[i]
            total = count[0] + count[1]
            fraction = count[0]/total if total >= min_depth else -1
            # mut_fractions[i].append(round(fraction, 2))
            mut_fractions[i].append(round(fraction, 4))

    mut_names = set()
    mut_names = [n for n in names]
    csv_headers = ['Mutation'] + [n + ' %' for n in sample_names]
    csv_rows = []
    for i in range(num_mutations):
        sm = mut_fractions[i]
        csv_rows.append([str(mut_names[i])] + [str(sm[j]) for j in range(len(sample_names))])
    with open('{0}_{1}_mutations.csv'.format(home_lab, mutants_name), 'w') as f:
        f.write('\n'.join(','.join(row) for row in [csv_headers] + csv_rows))

def find_mutants_in_bam(bam_path, mutations):
    import pysam

    samfile = pysam.Samfile(bam_path, "rb")

    # parsed_muts = [parse_snv(mut) for mut in mutations]
    parsed_muts = {}
    for mut in mutations:
        parsed_muts[mut] = parse_mutation(mut)
    mut_results = {mut: {snv_name(m): [0,0] for m in parsed_muts[mut]} for mut in mutations}

    for pileupcolumn in samfile.pileup(stepper='nofilter'):
        pos = pileupcolumn.pos + 1
        for mut in parsed_muts:
            for m in parsed_muts[mut]:
                if pos == m[1]:
                    muts, not_muts = mut_in_col(pileupcolumn, m[2])
                    mut_results[mut][snv_name(m)] = [muts, not_muts]
    samfile.close()

    for mut in mut_results:
        max_freq = -1
        max_muts = [0,0]
        for m in mut_results[mut]:
            muts, not_muts = mut_results[mut][m]
            freq = muts/(muts+not_muts) if muts + not_muts > 0 else 0
            if freq > max_freq:
                max_freq = freq
                max_muts = [muts, not_muts]
        mut_results[mut] = max_muts

    return mut_results


def mut_idx(mut):
    # Sort by genomic index of mutations
    snvs = parse_mutation(mut)
    if len(snvs) == 0:
        return -1
    return snvs[0][1]


# def find_mutants(file_path, mutations_path, min_depth, not_in): #TODO: not in lineage
def find_mutants(file_path, mutations_path, min_depth, save_img, csv):
    """
    Accepts either a bam file or a tab delimited  txt file like
    s1.bam  Sample 1
    s2.bam  Sample 2
    """

    sample_results = []
    sample_names = []
    home_lab = file_path.replace('.txt', '')
    lineages = list(mut_lins['C241T'].keys()) # arbitrary
    print(lineages)
    if mutations_path in lineages:
        lin = mutations_path
        print('Searcing for {} mutations'.format(lin))
        mutations = [mut for mut in mut_lins if mut_lins[mut][lin] > 0 and mut_idx(mut) != -1]
        # Unique
        # mutations = [mut for mut in mutations if all(mut_lins[mut][l] == 0 for l in lineages if l != lin)]
        # mutations = [mut for mut in mutations if sum(mut_lins[mut][l] for l in lineages) == 1]
        mutations.sort(key=mut_idx)
    else:
        print('Searching for mutations in {}'.format(mutations_path))
        with open(mutations_path, 'r') as f:
            mutations = [mut for mut in f.read().split('\n') if len(mut)]

    if file_path.endswith('.bam'):
        sample_results.append(find_mutants_in_bam(file_path, mutations))
        sample_names.append('')
    else:
        with open(file_path, 'r') as f:
            samples = [line.split('\t') for line in f.read().split('\n')]
        for sample in samples:
            if sample[0].endswith('.bam'): # Mostly for filtering empty
                print('{}:'.format(sample[1]))
                sample_results.append(find_mutants_in_bam(sample[0], mutations))
                sample_names.append(sample[1])
                print_mut_results(sample_results[-1], min_depth)
                print()

    mutants_name = mutations_path.replace('.txt', '').replace('.', '')
  #  img_path = file_path.replace('.bam', '_{}_mutants.png'.format(mutants_name)) if save_img else None
    if save_img:
        img_path = file_path.replace('.txt', '_{}_mutations.png'.format(mutants_name))
    else:
        img_path=None

    if csv:
        write_csv(sample_results, sample_names, min_depth, home_lab, mutants_name)

    plot_mutations(sample_results, sample_names, min_depth, img_path)



import re

from .sars_cov_2 import genes, seq


codons = { 
    'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M', 
    'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T', 
    'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K', 
    'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R',                  
    'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L', 
    'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P', 
    'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q', 
    'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R', 
    'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V', 
    'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A', 
    'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E', 
    'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G', 
    'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S', 
    'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L', 
    'TAC':'Y', 'TAT':'Y', 'TAA':'_', 'TAG':'_', 
    'TGC':'C', 'TGT':'C', 'TGA':'_', 'TGG':'W', 
} 

nts = 'ACGT'


def aa(mut):
    gene = mut[:mut.find(':')]
    if gene == 'DEL':
        nt_idx, length = re.findall(r'\d+', mut)
        nt_idx = int(nt_idx)
        length = int(length)
        return ['{}{}-'.format(seq[nt_idx+i], nt_idx+i) for i in range(length)]
    aa_idx = int(re.findall(r'\d+', mut)[-1])
    nt_idx = genes[gene][0] + (aa_idx - 1) * 3
    codon = seq[nt_idx:nt_idx+3]
    if mut[mut.find(':')+1:].startswith('DEL') or mut[-1] == '-':
        return ['{}{}-'.format(seq[nt_idx],nt_idx+1) for nt_idx in range(nt_idx,nt_idx+3)]
    new_acid = mut[-1]
    acid = codons[codon]
    nt_muts = []

    for i in range(4):
        if codons[nts[i] + codon[1] + codon[2]] == new_acid:
            nt_muts.append('{}{}{}'.format(codon[0], nt_idx + 1, nts[i]))

    for i in range(4):
        if codons[codon[0] + nts[i] + codon[2]] == new_acid:
            nt_muts.append('{}{}{}'.format(codon[1], nt_idx + 2, nts[i]))

    for i in range(4):
        if codons[codon[0] + codon[1] + nts[i]] == new_acid:
            nt_muts.append('{}{}{}'.format(codon[2], nt_idx + 3, nts[i]))

# if a single nt change cannot make this aa change, try all double nt changes
#    
#    if nt_muts == []:
#        for i in range(4):
#            for j in range(4):
#                if codons[nts[i] + nts[j] + codon[2]] == new_acid:
#                    nt_muts.append('{}{}{}'.format(codon[0], nt_idx + 1, nts[i]))
#                    nt_muts.append('{}{}{}'.format(codon[1], nt_idx + 2, nts[j]))
#        for i in range(4):
#            for j in range(4):
#                if codons[codon[0] + nts[i] + nts[j]] == new_acid:
#                    nt_muts.append('{}{}{}'.format(codon[1], nt_idx + 2, nts[i]))
#                    nt_muts.append('{}{}{}'.format(codon[2], nt_idx + 3, nts[j]))
#        for i in range(4):
#            for j in range(4):
#                if codons[nts[i] + codon[1] + nts[j]] == new_acid:
#                    nt_muts.append('{}{}{}'.format(codon[0], nt_idx + 1, nts[i]))
#                    nt_muts.append('{}{}{}'.format(codon[2], nt_idx + 3, nts[j]))

    return nt_muts


def nt(mut):
    base = mut[0]
    new_base = mut[-1]
    nt_idx = int(re.findall(r'\d+', mut)[0]) - 1
    for gene in genes:
        l = genes[gene]
        if l[0] < nt_idx and l[1] > nt_idx:
            nt_offset = (nt_idx - l[0]) % 3
            # Slightly roundabout way of getting start of codon
            aa_idx = int((nt_idx - l[0]) / 3) + 1
            if new_base == '-':
                return '{}:DEL{}'.format(gene, aa_idx)
            codon_start = l[0] + (aa_idx - 1) * 3
            codon = list(seq[codon_start:codon_start+3])
            acid = codons[''.join(codon)]
            codon[nt_offset] = new_base
            new_acid = codons[''.join(codon)]
            return '{}:{}{}{}'.format(gene, acid, aa_idx, new_acid)




from setuptools import setup

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='alcov',
    version='1.1.9',
    description='Identify frequencies of concerning mutations from aligned reads',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Isaac Ellmen',
    author_email='isaac.ellmen@uwaterloo.ca',
    maintainer='Jenn Knapp',
    maintainer_email='jenn.knapp@uwaterloo.ca',
    packages=['alcov'],
    url='https://github.com/Ellmen/alcov',
    install_requires=[
        'fire',
        'numpy',
        'pandas',
        'scikit-learn>=0.24',
        'matplotlib',
        'seaborn',
        'pysam',
    ],
    entry_points={
        'console_scripts': ['alcov=alcov.command_line:main'],
    }
)
